//---------------------------------------------------------------------------
#ifndef commonH
#define commonH
//---------------------------------------------------------------------------
#define PROG_NAME             "Quido"
#define PROG_EXE_NAME         "quido"
#define PROG_VERSION          0x14
#define PROG_AUTHOR           "RM-TEAM 04/2010, 09/2013, 10/2015, 08/2016"

//---------------------------------------------------------------------------
#define BYTE                  unsigned char
#define WORD                  unsigned short int
#define BOOL                  int
#define FALSE                 0
#define TRUE                  !FALSE

//---------------------------------------------------------------------------
#define PACK_MODE_1           1
#define PACK_MODE_2           2
#define PACK_MODE_3           4
#define PACK_MODE_4           8

#define MAX_FILE_LENGTH       65536
#define BUFFER_LENGTH         (MAX_FILE_LENGTH + 16384)

//---------------------------------------------------------------------------
#define ERR_PRINT_USAGE       -1
#define ERR_PRINT_VERSION     -2
#define ERR_OUT_OF_MEMORY     -3
#define ERR_BAD_OPTIONS       -4
#define ERR_BAD_PACK_MODE     -5
#define ERR_OPEN_SRC_FILE     -6
#define ERR_ACCESS_SRC_FILE   -7
#define ERR_LENGTH_SRC_FILE   -8
#define ERR_READ_SRC_FILE     -9
#define ERR_OPEN_DEST_FILE    -10
#define ERR_ACCESS_DEST_FILE  -11
#define ERR_WRITE_DEST_FILE   -12
#define ERR_BAD_COLOR_VALUE   -13
#define ERR_BAD_ADDRESS       -14
#define ERR_BAD_UNPACK_ADDR   -15
#define ERR_INEFFECTIVE       -16
#define ERR_NO_INPUT_FILE     -17
#define ERR_INPUT_FILE_MORE   -18
#define ERR_OUTPUT_FILE_MORE  -19
#define ERR_BAD_EXPAND_SIZE   -20
#define ERR_BAD_NAME          -21

//---------------------------------------------------------------------------
typedef struct {
  char *fnIn;         // pointer to the source file path/name
  char *fnOut;        // pointer to the destination file path/name
  char fn[256];       // destination file path/name
  char name[11];      // name of the tape block
  BOOL vb;            // verborse print
  BOOL append;        // append result to the output file
  BOOL i8080;         // insert i8080 unpack routine before packed data
  BOOL vram;          // locate unpack routine into VRAM (PMD 85)
  BOOL pmd;           // store output to the PTP file (PMD 85)
  BOOL z80;           // insert z80 standard unpack routine before packed data
  BOOL zx;            // store output to the TAP file (ZX Spectrum)
  BOOL bas;           // create BASIC loader (TAP file)
  BOOL d80;           // create D80 syntax in BASIC loader (TAP file)
  BOOL hp;            // include POKE in BASIC loader for disabling tape headers (TAP file)
  BOOL ei;            // expand input file by zeros if packing was ineffective
  BOOL ez;            // expand input file by zeros implicitly
  BOOL force;         // store original file if packing was ineffective
  BOOL prolog;        // insert AllRAM Prolog before unpack routine
  int em;             // expand input file max up to size bytes
  int useModes;       // requested packing modes
  int modeMask;       // used packing mode mask
  int mode;           // used packing mode
  int colors[3];      // colors in BASIC loader (TAP file)
  int clearAddr;      // addres for CLEAR in BASIC loader (TAP file)
  int runAddr;        // address where to start bin. file for BASIC loader
  int targetAddr;     // start address of binary file
  int unpackRtnAddr;  // unpack routine address
  int jumpAddr;       // jump address after unpack
  int originalLen;    // length of original data
  int packedLen[4];   // length of packed data
  int realDataLen[4]; // length of real data in packed block
  int overheadLen[4]; // length of intermediate data in packed block
  int moverLen[4];    // length of the Mover of the unpack routine
  int rtnLen[4];      // length of the unpack routine itself
  int totalLen[4];    // length of packed block including routine
  int prologSP;       // value of the SP in AllRAM Prolog
} WORK_PARAMS;

//---------------------------------------------------------------------------
// pack.c
int pack(BYTE *buffer, int length, int packMode,
         int *realDataLength, int *overheadLength);

// zxstape.c
void getRoutinesLengthZ80(WORK_PARAMS *wp);
void constructBasicZX(WORK_PARAMS *wp, BYTE **out);
void constructCodeZX(WORK_PARAMS *wp, BYTE **out, BYTE *packed);
void constructBinZ80(WORK_PARAMS *wp, BYTE **out, BYTE *packed);

// pmdtape.c
void getRoutinesLengthI8080(WORK_PARAMS *wp);
void constructBinI8080(WORK_PARAMS *wp, BYTE **out, BYTE *packed);
void constructCodePMD(WORK_PARAMS *wp, BYTE **out, BYTE *packed);

//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------

 