//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS *wp);
int parseNumber(char *str);
int readFile(char *fileName, BYTE *buff, int maxLen);
int writeFile(char *fileName, BYTE *buff, int len, BOOL append);
int printError(int err, void *par);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  WORK_PARAMS *wp = NULL;
  BYTE *buff1 = NULL, *buff2 = NULL, *pb;
  int i, err, bestLen, finalLen;
  char *pc;

  do {
    wp = (WORK_PARAMS *) malloc(sizeof(WORK_PARAMS));
    if (!wp) {
      err = ERR_OUT_OF_MEMORY;
      break;
    }

    // spracovanie a kontrola vstupnych parametrov
    err = processArguments(argc, argv, wp);
    if (err != 0)
      break;

    // ziskanie dlzok rozpakovacich rutin
    if (wp->z80)
      getRoutinesLengthZ80(wp);
    else if (wp->i8080)
      getRoutinesLengthI8080(wp);
    else {
      memset(wp->moverLen, 0, sizeof(wp->moverLen));
      memset(wp->rtnLen, 0, sizeof(wp->rtnLen));
    }

    // alokuj buffre
    buff1 = (BYTE *) malloc(BUFFER_LENGTH);
    if (!buff1) {
      err = ERR_OUT_OF_MEMORY;
      break;
    }

    buff2 = (BYTE *) malloc(BUFFER_LENGTH);
    if (!buff2) {
      err = ERR_OUT_OF_MEMORY;
      break;
    }

    // vycisti buffre
    memset(buff1, 0, BUFFER_LENGTH);
    memset(buff2, 0, BUFFER_LENGTH);

    // precitanie vstupneho suboru
    if (wp->vb)
      printf("Reading input file '%s'\n", wp->fnIn);
    wp->originalLen = readFile(wp->fnIn, buff1, MAX_FILE_LENGTH);
    if (wp->originalLen < 0) {
      err = wp->originalLen;
      break;
    }
    if (wp->vb)
      printf("Source data length: %d bytes\n", wp->originalLen);

    // rozsir subor nulami, ak je to pozadovane uz pred pakovanim (-ez)
    if (wp->ez && wp->originalLen < wp->em) {
      i = (wp->originalLen + 31) & ~31;
      if (i - wp->originalLen < 5)
        i += 16;
      if (i > wp->em)
        i = wp->em;
      if (i > MAX_FILE_LENGTH)
        i = MAX_FILE_LENGTH;
      wp->originalLen = i;
      if (wp->vb)
        printf("Source data length after expanding by zeros: %d bytes\n", wp->originalLen);
    }

    if (wp->vb)
      printf("\n");

    // ak je po neuspesnom pakovani pozadovane rozsirenie suboru nulami (-ei),
    // subor sa bude rozsirovat az po maximalne povolenu velkost (-em)
    while (TRUE) {
      // spakovanie dat zvolenymi modmi
      for (i = 0; i < 4; i++) {
        wp->mode = i;
        wp->modeMask = wp->useModes & (1 << i);
        wp->packedLen[i] = -1;
        wp->totalLen[i] = -1;
        if (wp->modeMask) {
          if (wp->vb)
            printf("Packing in mode: %d\n", i + 1);
          memcpy(buff2, buff1, BUFFER_LENGTH);
          wp->packedLen[i] = pack(buff2, wp->originalLen, wp->modeMask,
                                  &wp->realDataLen[i], &wp->overheadLen[i]);
          if (wp->packedLen[i] < 0) {
            if (wp->vb)
              printf("  Packing ineffective!\n\n");
            continue;
          }

          if (wp->z80 || wp->i8080)
            wp->totalLen[i] = wp->moverLen[i] + wp->rtnLen[i] + wp->packedLen[i];
          else
            wp->totalLen[i] = wp->packedLen[i];

          if (wp->vb) {
            printf("  Packed data length: %d bytes", wp->packedLen[i]);
            printf("  Data: %d  Overhead: %d\n", wp->realDataLen[i], wp->overheadLen[i]);
            if (wp->z80 || wp->i8080) {
              printf("  Unpack routine length: %d bytes, with mover: %d bytes", wp->rtnLen[i], wp->moverLen[i] + wp->rtnLen[i]);
              if (wp->vram)
                printf(" (%d VRAM blocks)", (wp->rtnLen[i] + 15) / 16);
              printf("\n");
            }
            printf("  Total length: %d bytes", wp->totalLen[i]);
            printf(" (%.2f%%)\n", (wp->totalLen[i] * 100.0) / wp->originalLen);
          }
          printf("\n");
        }
      }

      // najdenie najlepsieho vysledku
      bestLen = wp->originalLen;
      wp->mode = -1;
      for (i = 0; i < 4; i++) {
        if (wp->totalLen[i] > 0 && wp->totalLen[i] < bestLen) {
          bestLen = wp->totalLen[i];
          wp->mode = i;
        }
      }

      if (wp->mode >= 0
          || (wp->mode < 0 && ((wp->useModes & 3) == 0 || !wp->ei || wp->originalLen >= wp->em)))
        break;

      // rozsir subor nulami, ak je to pozadovane po neuspesnom pakovanim
      i = (wp->originalLen + 31) & ~31;
      if (i - wp->originalLen < 5)
        i += 16;
      if (i > wp->em)
        i = wp->em;
      if (i > MAX_FILE_LENGTH)
        i = MAX_FILE_LENGTH;
      wp->originalLen = i;
      if (wp->vb)
        printf("Source data length expanded by zeros after ineffective packing to %d bytes\n\n", wp->originalLen);
    }

    if (wp->mode < 0) {
      err = printError(ERR_INEFFECTIVE, NULL);
      if (!wp->force)
        break;
    }

    if (wp->mode >= 0) {
      wp->modeMask = 1 << wp->mode;
      // konecne pakovanie najlepsim modom
      memcpy(buff2, buff1, BUFFER_LENGTH);
      wp->packedLen[wp->mode] = pack(buff2, wp->originalLen, wp->modeMask,
                                     &wp->realDataLen[wp->mode], &wp->overheadLen[wp->mode]);
      if (wp->z80 || wp->i8080)
        wp->totalLen[wp->mode] = wp->moverLen[wp->mode] + wp->rtnLen[wp->mode] + wp->packedLen[wp->mode];
      else
        wp->totalLen[wp->mode] = wp->packedLen[wp->mode];

      printf("Packed in mode: %d\n", wp->mode + 1);
      printf("Original length: %d bytes\n", wp->originalLen);
      if (wp->vb) {
        printf("  Packed data length: %d bytes", wp->packedLen[wp->mode]);
        printf("  Data: %d  Overhead: %d\n", wp->realDataLen[wp->mode], wp->overheadLen[wp->mode]);
      }
      printf("New length: %d bytes (%.2f%%)\n\n", wp->totalLen[wp->mode], (wp->totalLen[wp->mode] * 100.0) / wp->originalLen);
    }
    else {
      wp->mode = 0;
      memcpy(buff2, buff1, BUFFER_LENGTH);
      wp->packedLen[wp->mode] = wp->originalLen;
      wp->totalLen[wp->mode] = wp->originalLen;
      printf("Source data cannot be packed and will be stored as is!\n");
    }

    finalLen = wp->totalLen[wp->mode];
    pb = buff1;
    if (wp->zx || wp->pmd) {
      if (wp->vb)
        printf("Constructing %s file\n", (wp->zx) ? "TAP" : "PTP");

     // priprava mena TAP/PTP bloku z nazvu zdrojoveho suboru (ak nebolo zadane)
      if (!*wp->name) {
        pc = strrchr(wp->fnIn, '/');
        if (pc == NULL)
          pc = strrchr(wp->fnIn, '\\');
        if (pc == NULL)
          pc = wp->fnIn;
        else
          pc++;
        i = (wp->zx) ? 10 : 8;
        strncpy(wp->name, pc, i);
        wp->name[i] = 0;
        pc = strrchr(wp->name, '.');
        if (pc != NULL)
          *pc = 0;
      }

      if (wp->zx) {
        // ZX Spectrum TAP
        if (wp->bas)
          constructBasicZX(wp, &pb);
        constructCodeZX(wp, &pb, buff2);
        finalLen = pb - buff1;
      }
      else { // if (wp->pmd)
        // PMD 85 PTP
        constructCodePMD(wp, &pb, buff2);
        finalLen = pb - buff1;
      }
    }
    else if (wp->z80 && wp->mode >= 0) {
      constructBinZ80(wp, &pb, buff2);
      finalLen = pb - buff1;
    }
    else if (wp->i8080 && wp->mode >= 0) {
      constructBinI8080(wp, &pb, buff2);
      finalLen = pb - buff1;
    }
    else
      memcpy(buff1, buff2, BUFFER_LENGTH);

    // zapis vystupneho suboru
    if (wp->vb)
      printf("Writing to the output file '%s'\n", wp->fnOut);
    err = writeFile(wp->fnOut, buff1, finalLen, wp->append);

  } while (FALSE);

  if (err == ERR_PRINT_USAGE) {
    usage();
    err = 0;
  }
  else if (err == ERR_PRINT_VERSION) {
    version();
    err = 0;
  }
  else if (!err)
    printf("\n");

  if (wp)
    free(wp);
  if (buff1)
    free(buff1);
  if (buff2)
    free(buff2);

#ifdef _DEBUG
  getchar();
#endif

  return -err;
}
//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS *wp)
{
  int err = 0, i;
  char *pc;

  // inicializacia datoveho bloku
  memset(wp, 0, sizeof(WORK_PARAMS));
  wp->colors[2] = 7; // default INK 7
  wp->clearAddr = -1;
  wp->runAddr = -1;
  wp->targetAddr = -1;
  wp->unpackRtnAddr = -1;
  wp->jumpAddr = -1;
  wp->em = -1;
  wp->useModes = (PACK_MODE_1 | PACK_MODE_2 | PACK_MODE_3 | PACK_MODE_4);
  wp->prologSP = -1;

  // spracovanie a kontrola vstupnych parametrov
  do {
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        err = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-v") == 0) {
        err = ERR_PRINT_VERSION;
        break;
      }
      else if (strcmp(argv[i], "-vb") == 0) {
        wp->vb = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-z80") == 0) {
        wp->z80 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-i8080") == 0) {
        wp->i8080 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ptp") == 0) {
        wp->pmd = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-vram") == 0) {
        wp->vram = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-tap") == 0) {
        wp->zx = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-b") == 0) {
        wp->bas = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-d80") == 0) {
        wp->d80 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-hp") == 0) {
        wp->hp = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-append") == 0) {
        wp->append = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ei") == 0) {
        wp->ei = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ez") == 0) {
        wp->ez = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-em") == 0) {
        if (++i == argc)
          wp->em = 0;
        else
          wp->em = parseNumber(argv[i]);
        if (wp->em <= 0 || wp->em > MAX_FILE_LENGTH) {
          err = printError(ERR_BAD_EXPAND_SIZE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-force") == 0) {
        wp->force = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-m") == 0) {
        wp->useModes = 0;
        if ((i + 1) < argc) {
          pc = argv[++i];
          while (*pc) {
            if (*pc < '1' || *pc > '4')
              break;
            if (wp->useModes & (1 << (*pc - '1')))
              break;
            wp->useModes |= (1 << (*pc - '1'));
            pc++;
          }
        }
        if (wp->useModes == 0 || *pc != 0) {
          err = printError(ERR_BAD_PACK_MODE, argv[i]);
          break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-cb") == 0) {
        if (++i == argc)
          wp->colors[0] = -1;
        else
          wp->colors[0] = parseNumber(argv[i]);
        if (wp->colors[0] < 0 || wp->colors[0] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-cp") == 0) {
        if (++i == argc)
          wp->colors[1] = -1;
        else
          wp->colors[1] = parseNumber(argv[i]);
        if (wp->colors[1] < 0 || wp->colors[1] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ci") == 0) {
        if (++i == argc)
          wp->colors[2] = -1;
        else
          wp->colors[2] = parseNumber(argv[i]);
        if (wp->colors[2] < 0 || wp->colors[2] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-a") == 0) {
        if (++i == argc)
          wp->targetAddr = -1;
        else
          wp->targetAddr = parseNumber(argv[i]);
        if (wp->targetAddr < 0 || wp->targetAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ra") == 0) {
        if (++i == argc)
          wp->unpackRtnAddr = -1;
        else
          wp->unpackRtnAddr = parseNumber(argv[i]);
        if (wp->unpackRtnAddr < 0 || wp->unpackRtnAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ja") == 0) {
        if (++i == argc)
          wp->jumpAddr = -1;
        else
          wp->jumpAddr = parseNumber(argv[i]);
        if (wp->jumpAddr < 0 || wp->jumpAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-c") == 0) {
        if (++i == argc)
          wp->clearAddr = -1;
        else
          wp->clearAddr = parseNumber(argv[i]);
        if (wp->clearAddr < 0 || wp->clearAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-r") == 0) {
        if (++i == argc)
          wp->runAddr = -1;
        else
          wp->runAddr = parseNumber(argv[i]);
        if (wp->runAddr < 0 || wp->runAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (++i == argc || wp->fnOut != NULL) {
          err = printError(ERR_OUTPUT_FILE_MORE, NULL);
          break;
        }
        wp->fnOut = argv[i];
        continue;
      }
      else if (strcmp(argv[i], "-n") == 0) {
        if (++i == argc)
          wp->name[0] = '\0';
        else {
          if (strlen(argv[i]) > 10) {
            err = printError(ERR_BAD_NAME, argv[i]);
            break;
          }
          else
            strcpy(wp->name, argv[i]);
        }
      }
      else if (strcmp(argv[i], "-p") == 0) {
        wp->prolog = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-sp") == 0) {
        if (++i == argc)
          wp->prologSP = -1;
        else
          wp->prologSP = parseNumber(argv[i]);
        if (wp->prologSP < 0 || wp->prologSP > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else {
        if (wp->fnIn != NULL) {
          err = printError(ERR_INPUT_FILE_MORE, NULL);
          break;
        }
        wp->fnIn = argv[i];
        continue;
      }
    }
    if (err != 0)
      break;

    if (wp->fnIn == NULL) {
      err = printError(ERR_NO_INPUT_FILE, NULL);
      break;
    }

    // kontrola povolenej kombinacie volieb
    if ((wp->zx && (wp->pmd || wp->i8080))
        || (wp->z80 && wp->i8080)
        || ((wp->vram || wp->prolog) && (!wp->i8080 || wp->z80 || wp->zx))
        || (wp->pmd && (wp->z80 || wp->bas || wp->hp || wp->d80 || wp->runAddr >= 0 || wp->clearAddr >= 0))
        || (!wp->zx && (wp->bas || wp->hp || wp->d80 || wp->runAddr >= 0 || wp->clearAddr >= 0))
        /* || (wp->ei && wp->ez) */) {
      err = printError(ERR_BAD_OPTIONS, NULL);
      break;
    }

    if (*wp->name && wp->pmd) {
      if (strlen(wp->name) > 8) {
        err = printError(ERR_BAD_NAME, wp->name);
        break;
      }
    }

    // priprava mena vystupneho suboru, ak nebol zadany
    if (wp->fnOut == NULL) {
      strcpy(wp->fn, wp->fnIn);
      pc = strrchr(wp->fn, '.');
      if (pc != NULL)
        *pc = 0;
      if (wp->zx)
        strcat(wp->fn, ".tap");
      else if (wp->pmd)
        strcat(wp->fn, ".ptp");
      else
        strcat(wp->fn, ".pck");
      wp->fnOut = wp->fn;
      if (strcmp(wp->fnOut, wp->fnIn) == 0)
        strcat(wp->fnOut, ".pck");
    }

    // default hodnoty
    if (wp->z80 || wp->i8080) {
      if (wp->targetAddr < 0)
        wp->targetAddr = ((wp->z80) ? 32768 : 0);
      if (wp->unpackRtnAddr < 0)
        wp->unpackRtnAddr = ((wp->z80) ? 23456 : 0x7F00);
      if (wp->i8080 && wp->vram) {
        if (wp->unpackRtnAddr < 0xC030 || wp->unpackRtnAddr > 0xFDF0
            || (wp->unpackRtnAddr & 0x3F) != 0x30) {
          err = printError(ERR_BAD_UNPACK_ADDR, &wp->unpackRtnAddr);
          break;
        }
      }
    }
    if (wp->zx && wp->runAddr < 0)
      wp->runAddr = 32768;
    if (wp->zx && wp->clearAddr < 0)
      wp->clearAddr = 24575;
    if (wp->i8080 && wp->prologSP < 0)
      wp->prologSP = 49152;
    if (wp->em < 0)
      wp->em = MAX_FILE_LENGTH;

  } while (FALSE);

  return err;
}
//---------------------------------------------------------------------------
int parseNumber(char *str)
{
  char *p;
  int val = strtol(str, &p, 0);
  if (val < 0 || (val == 0 && *p != 0))
    val = -1;
  return val;
}
//---------------------------------------------------------------------------
int readFile(char *fileName, BYTE *buff, int maxLen)
{
  int err = 0, len;
  FILE *f = NULL;

  do {
    f = fopen(fileName, "rb");
    if (f == NULL) {
      err = printError(ERR_OPEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    len = ftell(f);
    if (len < 0 || len > maxLen) {
      err = printError(ERR_LENGTH_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    if (fread(buff, 1, len, f) != (unsigned int) len) {
      err = printError(ERR_READ_SRC_FILE, fileName);
      break;
    }
  } while (FALSE);

  if (f != NULL)
    fclose(f);

  if (err < 0)
    return err;

  return len;
}
//---------------------------------------------------------------------------
int writeFile(char *fileName, BYTE *buff, int len, BOOL append)
{
  int err = 0;
  FILE *f = NULL;

  do {
    if (append)
      f = fopen(fileName, "ab");
    else
      f = fopen(fileName, "wb");
    if (f == NULL) {
      err = printError(ERR_OPEN_DEST_FILE, fileName);
      break;
    }
    if (!append && fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_DEST_FILE, fileName);
      break;
    }
    if (fwrite(buff, 1, len, f) != (unsigned int) len) {
      err = printError(ERR_WRITE_DEST_FILE, fileName);
      break;
    }
  } while (FALSE);

  if (f != NULL)
    fclose(f);

  return err;
}
//---------------------------------------------------------------------------
int printError(int err, void *par)
{
  switch (err) {
    case ERR_OUT_OF_MEMORY :
      fprintf(stderr, "Out of memory!");
      break;

    case ERR_BAD_PACK_MODE :
      fprintf(stderr, "Bad packing mode specified: '%s'!", (char *) par);
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of options!");
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open input file '%s'!", (char *) par);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access input file '%s'!", (char *) par);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the input file '%s'!", (char *) par);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the input file '%s'!", (char *) par);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *) par);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *) par);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *) par);
      break;

    case ERR_BAD_COLOR_VALUE :
      fprintf(stderr, "After '%s' must follow color in range 0 to 7!", (char *) par);
      break;

    case ERR_BAD_ADDRESS :
      fprintf(stderr, "After '%s' must follow address in range 0 to 65535!", (char *) par);
      break;

    case ERR_BAD_EXPAND_SIZE :
      fprintf(stderr, "After '%s' must follow max size of the file after expanding! (max %d)", (char *) par, MAX_FILE_LENGTH);
      break;

    case ERR_BAD_UNPACK_ADDR :
      fprintf(stderr, "Bad address of the unpack routine on PMD 85: 0x%04X\n", *(int *) par);
      fprintf(stderr, "Address of the VRAM located unpack routine must satisfy following condition:\n");
      fprintf(stderr, "  addr >= 0xC030 AND addr <= 0xFDF0 AND (addr & 0xC03F) = 0xC030");
      break;

    case ERR_INEFFECTIVE :
      fprintf(stderr, "Packing ineffective!");
      break;

    case ERR_NO_INPUT_FILE :
      fprintf(stderr, "Input file not specified!\nTry '%s -h' for help.", PROG_EXE_NAME);
      break;

    case ERR_INPUT_FILE_MORE :
      fprintf(stderr, "Only one input file can be specified!");
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified!");
      break;

    case ERR_BAD_NAME :
      fprintf(stderr, "Bad tape block name '%s'!", (char *) par);
      break;
  }

  fprintf(stderr, "\n\n");

  return err;
}
//---------------------------------------------------------------------------
void version(void)
{
  printf("%s v%d.%d - %s\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void pressEnter(void)
{
  printf("\nPress ENTER ...");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
void usage(void)
{
  version();

  printf("Usage: %s [options] infile.ext\n", PROG_EXE_NAME);
  printf("\nCommon options:\n");
  printf("  -o outfile.ext  output file name\n");
  printf("  -append\t  append at the end of the output file\n");
  printf("  -m 1234\t  pack only with given modes [all modes are trying]\n");
  printf("  -z80\t\t  insert Z80 unpack routine\n");
  printf("  -i8080\t  insert i8080 unpack routine\n");
  printf("  -a address\t  start address of binary file [32768 (z80) / 0 (i8080)]\n");
  printf("  -ra address\t  unpack routine address [23456 (z80) / 32512 (i8080)]\n");
  printf("  -ja address\t  jump address after unpack [ret]\n");
  printf("  -ei\t\t  expand input data by zeros on ineffective packing (modes 12)\n");
  printf("  -ez\t\t  expand input data by zeros implicitly\n");
  printf("  -em size\t  expand input data max up to size bytes [%d]\n", MAX_FILE_LENGTH);
  printf("  -force\t  store original data if packing was ineffective\n");
  printf("  -vb\t\t  verbose\n");
  printf("  -h\t\t  usage information\n");
  printf("  -v\t\t  version info\n");

  pressEnter();

  printf("\nOptions for PTP file / PMD 85:\n");
  printf("  -ptp\t\t  create PTP file\n");
  printf("  -n name\t  name of the tape block (max. 8 characters)\n");
  printf("  -p\t\t  insert AllRAM Prolog before unpack routine\n");
  printf("  -sp address\t  address of the SP register in AllRAM Prolog [49152]\n");
  printf("  -vram\t\t  locate unpack routine to the VideoRAM\n");

  printf("\nOptions for TAP file (ZX Spectrum):\n");
  printf("  -tap\t\t  create TAP file\n");
  printf("  -n name\t  name of the tape block (max. 10 characters)\n");
  printf("  -b\t\t  include BASIC loader\n");
  printf("  -c address\t  CLEAR address in BASIC loader [24575]\n");
  printf("  -r address\t  address where to start bin. file for BASIC loader [32768]\n");
  printf("  -cb colour\t  border color set by loader [0]\n");
  printf("  -cp colour\t  paper color set by loader [0]\n");
  printf("  -ci colour\t  ink color set by loader [7]\n");
  printf("  -d80\t\t  create D80 syntax loader\n");
  printf("  -hp\t\t  include POKE command for disabling tape headers\n");

  pressEnter();

  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");

  return;
}
//---------------------------------------------------------------------------

