//---------------------------------------------------------------------------
#include "common.h"
//---------------------------------------------------------------------------
#define IM1_DICT_SIZE         4096
#define IM1_PHRASE_MIN        3
#define IM1_PHRASE_MAX        (IM1_PHRASE_MIN + 7)        // 10
#define IM1_PHRASE_SHIFT      4
#define IM1_PHRASE_MASK       0x70
#define IM1_OFFSET_SHIFT      0
#define IM1_OFFSET_MASK       0x0F

#define IM2_DICT_SIZE         2048
#define IM2_PHRASE_MIN        3
#define IM2_PHRASE_MAX        (IM2_PHRASE_MIN + 15)       // 18
#define IM2_PHRASE_SHIFT      0
#define IM2_PHRASE_MASK       0x0F
#define IM2_OFFSET_SHIFT      4
#define IM2_OFFSET_MASK       0x70

#define SHRINK_MIN            3
#define SHRINK_MAX_LITTLE     (SHRINK_MIN + 63)           // 66
#define SHRINK_MAX_BIG        (SHRINK_MAX_LITTLE + 256)   // 322

#define NOT_PACKED_MAX        63

#define FLAG_IMPLOD           0x00    // 0LLLAAAA alebo 0AAALLLL
#define FLAG_SHRINK_LITTLE    0xC0    // 11LLLLLL
#define FLAG_SHRINK_BIG       0x80    // 10000000
#define FLAG_NO_PACKED        0x80    // 10LLLLLL

//---------------------------------------------------------------------------
int packBackward(void);
int pbFindBetter(BYTE *shrinkByte, int *implOff);
int pbShrink(BYTE *shrinkByte);
int pbImplod(int *implOff);

int packForward(void);
int pfFindBetter(BYTE *shrinkByte, int *implOff);
int pfShrink(BYTE *shrinkByte);
int pfImplod(int *implOff);

//---------------------------------------------------------------------------
// globalne premenne pre "pack.c"
int dictSize, phraseMin, phraseMax, phraseShift, phraseMask, offsetShift, offsetMask;
BYTE *pFirst, *pLast, *pAct;
int realDataLen, overheadLen;

//---------------------------------------------------------------------------
int pack(BYTE *buffer, int length, int packMode,
         int *realDataLength, int *overheadLength)
{
  int outLen;

  switch (packMode) {
    case PACK_MODE_1 :
    case PACK_MODE_3 :
      dictSize = IM1_DICT_SIZE;
      phraseMin = IM1_PHRASE_MIN;
      phraseMax = IM1_PHRASE_MAX;
      phraseShift = IM1_PHRASE_SHIFT;
      phraseMask = IM1_PHRASE_MASK;
      offsetShift = IM1_OFFSET_SHIFT;
      offsetMask = IM1_OFFSET_MASK;
      break;

    case PACK_MODE_2 :
    case PACK_MODE_4 :
      dictSize = IM2_DICT_SIZE;
      phraseMin = IM2_PHRASE_MIN;
      phraseMax = IM2_PHRASE_MAX;
      phraseShift = IM2_PHRASE_SHIFT;
      phraseMask = IM2_PHRASE_MASK;
      offsetShift = IM2_OFFSET_SHIFT;
      offsetMask = IM2_OFFSET_MASK;
      break;

    default :
      return ERR_BAD_PACK_MODE;
  }

  pFirst = buffer;
  pLast = buffer + length - 1;
  realDataLen = 0;
  overheadLen = 0;

  switch (packMode) {
    case PACK_MODE_1 :
    case PACK_MODE_2 :
      outLen = packBackward();
      break;

    case PACK_MODE_3 :
    case PACK_MODE_4 :
      outLen = packForward();
      break;
  }

  if (realDataLength)
    *realDataLength = realDataLen;
  if (overheadLength)
    *overheadLength = overheadLen;

  return outLen;
}
//---------------------------------------------------------------------------
// prevedie spakovanie bloku dat odzadu a vrati jeho novu dlzku. Spakovany blok
// je v povodnom buffri. Ak sa pakovanie nepodari, vrati sa chyba ERR_INEFFECTIVE.
int packBackward(void)
{
  int noPack, implOffset, cnt;
  BYTE *pTarget, shrByte;

  pAct = pLast; // ideme odzadu
  pTarget = pLast;
  noPack = 0;
  while (pAct >= pFirst) {
    cnt = pbFindBetter(&shrByte, &implOffset);
    if (cnt < 0) {      // nepakovatelne
      *pTarget-- = *pAct--;
      noPack++;
      if (noPack == 63) {
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget-- = (BYTE) (FLAG_NO_PACKED | noPack);
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        realDataLen += noPack;
        overheadLen++;
        noPack = 0;
      }
    }
    else {  // pakovatelne
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget-- = (BYTE) (FLAG_NO_PACKED | noPack);
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        realDataLen += noPack;
        overheadLen++;
        noPack = 0;
      }

      pAct -= cnt;  // posun aktualnej adresy
      if (implOffset >= 0) {  // Imploding
        *pTarget-- = (BYTE) implOffset;
        *pTarget-- = (BYTE) ((((cnt - phraseMin) << phraseShift) & phraseMask)
                             | (((implOffset / 256) << offsetShift) & offsetMask));
        overheadLen += 2;
      }
      else if (cnt > SHRINK_MAX_LITTLE) { // dlhy Shrink
        *pTarget-- = shrByte;
        *pTarget-- = (BYTE) (cnt - SHRINK_MAX_LITTLE - 1);
        *pTarget-- = (BYTE) FLAG_SHRINK_BIG;
        realDataLen++;
        overheadLen += 2;
      }
      else {  // kratky Shrink
        *pTarget-- = shrByte;
        *pTarget-- = (BYTE) (FLAG_SHRINK_LITTLE | (cnt - SHRINK_MIN));
        realDataLen++;
        overheadLen++;
      }
    }
  }

  if (noPack > 0) { // ukonci pripadny nepakovatelny blok
    *pTarget-- = (BYTE) (FLAG_NO_PACKED | noPack);
    realDataLen += noPack;
    overheadLen++;
  }

  cnt = pLast - pTarget;      // vysledna dlzka spakovaneho bloku
  while (pTarget < pLast)     // presun spakovany blok na zaciatok buffra
    *pFirst++ = *++pTarget;

  return cnt;
}
//---------------------------------------------------------------------------
// Pokusi sa najst na aktualnej pozicii pakovatelnu sekvenciu. Pokial sa takato
// sekvencia nenajde, vrati -1. Pokial sa najde, vrati jej dlzku. Ak sa jednalo
// o Imploding, na adresu implOff je ulozeny offset, kde sa nachadza dana fraza.
// Ak sa jedna o Shrinking, na adresu implOff je ulozena hodnota -1 a na adresu
// shrinkByte je ulozeny opakujuci sa byte.
int pbFindBetter(BYTE *shrinkByte, int *implOff)
{
  int shrCnt, implCnt, implOffset;

  *implOff = -1;

  shrCnt = pbShrink(shrinkByte); // skus Shrink
  if (shrCnt >= phraseMax)       // ak je pocet vacsi ako maximalna dlzka frazy,
    return shrCnt;               // nerob Implod a vrat sa rovno s tymto poctom

  implCnt = pbImplod(&implOffset);  // skus Implod
  if (implCnt > shrCnt) {           // ak je impolod uspesnejsi,
    *implOff = implOffset;          // nastav offset ku najdenej fraze
    return implCnt;                 // vrat sa s dlzkou frazy
  }

  return shrCnt;  // inak sa vrat s hodnotou so Shrinku, to moze byt aj -1
}
//---------------------------------------------------------------------------
// Vyhlada na aktualnej pozicii minimalne SHRINK_MIN a maximalne SHRINK_MAX_BIG
// rovnakych bytov a vrati ich pocet. Opakujuci sa byte je ulozeny na adresu
// shrinkByte. Pokial sa rovnake byty nenajdu, vrati -1.
int pbShrink(BYTE *shrinkByte)
{
  int cnt = -1;
  BYTE val, *p;

  if ((pAct - pFirst + 1) >= SHRINK_MIN
      && *pAct == *(pAct - 1) && *(pAct - 1) == *(pAct - 2)) { // SHRINK_MIN = 3
    cnt = SHRINK_MIN;
    val = *pAct;
    p = pAct - 3;
    while (p >= pFirst && cnt < SHRINK_MAX_BIG) {
      if (*p != val)
        break;
      p--;
      cnt++;
    }
    *shrinkByte = val;
  }

  return cnt;
}
//---------------------------------------------------------------------------
// Od aktualnej pozicie vyhlada frazu, ktora sa nachadza dalej k nizsim adresam.
// Najdena fraza musi mat minimalnu dlzku phraseMin a moze mat maximalne phraseMax.
// Ak je fraza najdena, na adresu implOff sa ulozi offset k cielovej fraze k nizsim
// adresam a vratena je dlzka tejto frazy. Pokial sa fraza nenajde, vrati sa -1.
int pbImplod(int *implOff)
{
  int cnt, cntD, implOffset, cntP, cntC;
  BYTE val, *p, *p1, *p2;

  cnt = -1;
  cntD = pAct - pFirst; // vypocitaj maximalnu velkost "okna" pre hladanie frazy
  if (cntD >= phraseMin) {
    if (cntD >= (dictSize + phraseMax)) // maximalne vsak do velkosti "slovnika"
      cntD = (dictSize + phraseMax);    //  vratane max. velkosti frazy

    p = pAct;     // adresa zdroja
    val = *p--;   // prvy byte frazy

    while (TRUE) {
      while (--cntD > 0 && *p-- != val) // hladaj prvy zhodny byte frazy
        ;
      if (cntD < cnt) // ak je zostavajuca velkost potencialnej frazy kratsia
        break;        // ako najvacsia najdena, rovno to ukonci, nema to zmysel

      implOffset = pAct - p - 1;  // vypocitaj offset k najdenemu zhodnemu bytu
      if (implOffset >= dictSize) // ak offset prekracuje velkost "okna",
        break;                    // ukonci to

      p1 = p;                     // hladaj dalsie rovnake byty
      p2 = pAct - 1;
      cntC = phraseMax - 1;       // maximalne vsak do maxima frazy alebo zostatku
      if (cntD < cntC)
        cntC = cntD;
      cntP = cntC + 1;
      while (cntC > 0) {          // hladaj frazu
        if (*p1 != *p2)
          break;
        p1--;
        p2--;
        cntC--;
      }
      cntC = cntP - cntC;                     // dlzka najdenej frazy
      if (cntC >= phraseMin && cntC >= cnt) { // ak je najdena fraza dostatocne
        cnt = cntC;                           // velka, akceptuj ju a zapamataj
        *implOff = implOffset;
      }

      if (cnt == cntP)  // ak si na konci "okna" a vacsia fraza uz nebude,
        break;          // ukonci to
    }
  }

  return cnt;
}
//---------------------------------------------------------------------------
//***************************************************************************
//---------------------------------------------------------------------------
// Prevedie spakovanie bloku dat spredu a vrati jeho novu dlzku. Spakovany blok
// je v povodnom buffri. Ak sa pakovanie nepodari, vrati sa chyba ERR_INEFFECTIVE.
int packForward(void)
{
  int noPack, implOffset, cnt;
  BYTE *pTarget, shrByte;

  pAct = pFirst; // ideme spredu
  pTarget = pFirst;
  noPack = 0;
  while (pAct <= pLast) {
    cnt = pfFindBetter(&shrByte, &implOffset);
    if (cnt < 0) {      // nepakovatelne
      *pTarget++ = *pAct++;
      noPack++;
      if (noPack == 63) {
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget++ = (BYTE) (FLAG_NO_PACKED | noPack);
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        realDataLen += noPack;
        overheadLen++;
        noPack = 0;
      }
    }
    else {  // pakovatelne
      if (noPack > 0) { // ukonci pripadny nepakovatelny blok
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        *pTarget++ = (BYTE) (FLAG_NO_PACKED | noPack);
        if (pTarget == pAct) // problem - data sa nedaju "rozumne" spakovat
          return ERR_INEFFECTIVE;
        realDataLen += noPack;
        overheadLen++;
        noPack = 0;
      }

      pAct += cnt;  // posun aktualnej adresy
      if (implOffset >= 0) {  // Imploding
        *pTarget++ = (BYTE) implOffset;
        *pTarget++ = (BYTE) ((((cnt - phraseMin) << phraseShift) & phraseMask)
                             | (((implOffset / 256) << offsetShift) & offsetMask));
        overheadLen += 2;
      }
      else if (cnt > SHRINK_MAX_LITTLE) { // dlhy Shrink
        *pTarget++ = shrByte;
        *pTarget++ = (BYTE) (cnt - SHRINK_MAX_LITTLE - 1);
        *pTarget++ = (BYTE) FLAG_SHRINK_BIG;
        realDataLen++;
        overheadLen += 2;
      }
      else {  // kratky Shrink
        *pTarget++ = shrByte;
        *pTarget++ = (BYTE) (FLAG_SHRINK_LITTLE | (cnt - SHRINK_MIN));
        realDataLen++;
        overheadLen++;
      }
    }
  }

  if (noPack > 0) { // ukonci pripadny nepakovatelny blok
    *pTarget++ = (BYTE) (FLAG_NO_PACKED | noPack);
    realDataLen += noPack;
    overheadLen++;
  }

  cnt = pTarget - pFirst;

  return cnt;
}
//---------------------------------------------------------------------------
// Pokusi sa najst na aktualnej pozicii pakovatelnu sekvenciu. Pokial sa takato
// sekvencia nenajde, vrati -1. Pokial sa najde, vrati jej dlzku. Ak sa jednalo
// o Imploding, na adresu implOff je ulozeny offset, kde sa nachadza dana fraza.
// Ak sa jedna o Shrinking, na adresu implOff je ulozena hodnota -1 a na adresu
// shrinkByte je ulozeny opakujuci sa byte.
int pfFindBetter(BYTE *shrinkByte, int *implOff)
{
  int shrCnt, implCnt, implOffset;

  *implOff = -1;

  shrCnt = pfShrink(shrinkByte); // skus Shrink
  if (shrCnt >= phraseMax)       // ak je pocet vacsi ako maximalna dlzka frazy,
    return shrCnt;               // nerob Implod a vrat sa rovno s tymto poctom

  implCnt = pfImplod(&implOffset);  // skus Implod
  if (implCnt > shrCnt) {           // ak je impolod uspesnejsi,
    *implOff = implOffset;          // nastav offset ku najdenej fraze
    return implCnt;                 // vrat sa s dlzkou frazy
  }

  return shrCnt;  // inak sa vrat s hodnotou so Shrinku, to moze byt aj -1
}
//---------------------------------------------------------------------------
// Vyhlada na aktualnej pozicii minimalne SHRINK_MIN a maximalne SHRINK_MAX_BIG
// rovnakych bytov a vrati ich pocet. Opakovany byte je ulozeny na adresu shrinkByte.
// Pokial sa rovnake byty nenajdu, vrati -1.
int pfShrink(BYTE *shrinkByte)
{
  int cnt = -1;
  BYTE val, *p;

  if ((pLast - pAct + 1) >= SHRINK_MIN
      && *pAct == *(pAct + 1) && *(pAct + 1) == *(pAct + 2)) { // SHRINK_MIN = 3
    cnt = SHRINK_MIN;
    val = *pAct;
    p = pAct + 3;
    while (p <= pLast && cnt < SHRINK_MAX_BIG) {
      if (*p != val)
        break;
      p++;
      cnt++;
    }
    *shrinkByte = val;
  }

  return cnt;
}
//---------------------------------------------------------------------------
// Od aktualnej pozicie vyhlada frazu, ktora sa nachadza dalej k vyssim adresam.
// Najdena fraza musi mat minimalnu dlzku phraseMin a moze mat maximalne phraseMax.
// Ak je fraza najdena, na adresu implOff sa ulozi offset k cielovej fraze k vyssim
// adresam a vratena je dlzka tejto frazy. Pokial sa fraza nenajde, vrati sa -1.
int pfImplod(int *implOff)
{
  int cnt, cntD, implOffset, cntP, cntC;
  BYTE val, *p, *p1, *p2;

  cnt = -1;
  cntD = pLast - pAct; // vypocitaj maximalnu velkost "okna" pre hladanie frazy
  if (cntD >= phraseMin) {
    if (cntD > (dictSize + phraseMax))  // maximalne vsak do velkosti "slovnika"
      cntD = (dictSize + phraseMax);    //  vratane max. velkosti frazy

    p = pAct;     // adresa zdroja
    val = *p++;   // prvy byte frazy

    while (TRUE) {
      while (--cntD > 0 && *p++ != val) // hladaj prvy zhodny byte frazy
        ;
      if (cntD < cnt) // ak je zostavajuca velkost potencialnej frazy kratsia
        break;        // ako najvacsia najdena, rovno to ukonci, nema to zmysel

      implOffset = p - pAct - 1;  // vypocitaj offset k najdenemu zhodnemu bytu
      if (implOffset >= dictSize) // ak offset prekracuje velkost "okna",
        break;                    // ukonci to

      p1 = p;                     // hladaj dalsie rovnake byty
      p2 = pAct + 1;
      cntC = phraseMax - 1;       // maximalne vsak do maxima frazy alebo zostatku
      if (cntD < cntC)
        cntC = cntD;
      cntP = cntC + 1;
      while (cntC > 0) {          // hladaj frazu
        if (*p1 != *p2)
          break;
        p1++;
        p2++;
        cntC--;
      }
      cntC = cntP - cntC;                     // dlzka najdenej frazy
      if (cntC >= phraseMin && cntC >= cnt) { // ak je najdena fraza dostatocne
        cnt = cntC;                           // velka, akceptuj ju a zapamataj
        *implOff = implOffset;
      }

      if (cnt == cntP)  // ak si na konci "okna" a vacsia fraza uz nebude,
        break;          // ukonci to
    }
  }

  return cnt;
}
//---------------------------------------------------------------------------

